@@ -1,5 +1,5 @@
 from abc import ABC, abstractmethod
-from dataclasses import dataclass, field
+from collections import deque
 from datetime import datetime, timezone
 from enum import Enum
 from functools import cached_property
@@ -12,6 +12,7 @@ from pathlib import Path
 import re
 from shutil import rmtree
 import subprocess
+import tempfile
 from typing import Any, Dict, Iterator, List, Match, Optional, Pattern, Tuple
 from urllib.parse import quote
 from zipfile import ZipFile
@@ -21,6 +22,7 @@ from click_loglevel import LogLevel
 from datalad.api import Dataset
 from dateutil.parser import isoparse
 from github import Github
+from github.Repository import Repository
 from github.Workflow import Workflow
 from github.WorkflowRun import WorkflowRun
 from in_place import InPlace
@@ -28,6 +30,8 @@ from pydantic import BaseModel, Field, validator
 import requests
 from yaml import safe_load
 
+from .util import expand_template
+
 log = logging.getLogger("tinuous")
 
 COMMON_STATUS_MAP = {
@@ -37,6 +41,7 @@ COMMON_STATUS_MAP = {
     "failed": "failed",
     "errored": "errored",
     "timed_out": "errored",
+    "startup_failure": "errored",
     "neutral": "incomplete",
     "action_required": "incomplete",
     "cancelled": "incomplete",
@@ -71,20 +76,22 @@ class EventType(Enum):
         }.get(travis_event)
 
 
-# We can't make this an ABC due to <https://github.com/python/mypy/issues/5374>
-@dataclass
-class CISystem:
+class CISystem(ABC, BaseModel):
     repo: str
-    token: str = field(repr=False)
+    token: str
     since: datetime
-    fetched: List[Tuple[datetime, bool]] = field(init=False, default_factory=list)
+    fetched: List[Tuple[datetime, bool]] = Field(default_factory=list)
 
     @staticmethod
+    @abstractmethod
     def get_auth_token() -> str:
-        raise NotImplementedError
+        ...
 
-    def get_build_logs(self, event_types: List[EventType]) -> Iterator["BuildLog"]:
-        raise NotImplementedError
+    @abstractmethod
+    def get_assets(
+        self, event_types: List[EventType], artifacts: bool = False
+    ) -> Iterator["Asset"]:
+        ...
 
     def register_build(self, ts: datetime, processed: bool) -> None:
         heapq.heappush(self.fetched, (ts, processed))
@@ -98,10 +105,13 @@ class CISystem:
             prev_ts = ts
         return prev_ts
 
+    class Config:
+        # <https://github.com/samuelcolvin/pydantic/issues/1241>
+        arbitrary_types_allowed = True
+        keep_untouched = (cached_property,)
 
-# We can't make this an ABC due to <https://github.com/python/mypy/issues/5374>
-@dataclass
-class BuildLog:
+
+class Asset(ABC, BaseModel):
     created_at: datetime
     event_type: EventType
     event_id: str
@@ -121,21 +131,30 @@ class BuildLog:
             "type": self.event_type.value,
             "type_id": self.event_id,
             "commit": self.commit,
+            "abbrev_commit": self.commit[:7],
             "number": str(self.number),
             "status": self.status,
             "common_status": COMMON_STATUS_MAP[self.status],
         }
 
-    def expand_path(self, path_template: str) -> str:
-        return path_template.format_map(self.path_fields())
+    def expand_path(self, path_template: str, vars: Dict[str, str]) -> str:
+        return expand_template(path_template, self.path_fields(), vars)
 
+    @abstractmethod
     def download(self, path: Path) -> List[Path]:
-        raise NotImplementedError
+        ...
+
+
+class BuildLog(Asset):
+    pass
+
+
+class Artifact(Asset):
+    pass
 
 
-@dataclass
 class GitHubActions(CISystem):
-    workflows: List[str]
+    workflows: Optional[List[str]] = None
 
     @staticmethod
     def get_auth_token() -> str:
@@ -160,18 +179,32 @@ class GitHubActions(CISystem):
 
     @cached_property
     def dl_session(self) -> requests.Session:
+        """
+        `requests.Session` used for downloading resources and other actions not
+        supported by Pygithub
+        """
         s = requests.Session()
         s.headers["Authorization"] = f"token {self.token}"
         return s
 
-    def get_build_logs(self, event_types: List[EventType]) -> Iterator["GHABuildLog"]:
+    @cached_property
+    def ghrepo(self) -> Repository:
+        return self.client.get_repo(self.repo)
+
+    def get_workflows(self) -> Iterator[Workflow]:
+        if self.workflows is None:
+            yield from self.ghrepo.get_workflows()
+        else:
+            for wffile in self.workflows:
+                yield self.ghrepo.get_workflow(wffile)
+
+    def get_assets(
+        self, event_types: List[EventType], artifacts: bool = False
+    ) -> Iterator["Asset"]:
         log.info("Fetching runs newer than %s", self.since)
-        repo = self.client.get_repo(self.repo)
-        for wffile in self.workflows:
-            wf = repo.get_workflow(wffile)
-            log.info("Fetching runs for workflow %s (%s)", wffile, wf.name)
-            for run in wf.get_runs():  # type: ignore[call-arg]
-                # <https://github.com/PyGithub/PyGithub/pull/1857>
+        for wf in self.get_workflows():
+            log.info("Fetching runs for workflow %s (%s)", wf.path, wf.name)
+            for run in wf.get_runs():
                 run_event = EventType.from_gh_event(run.event)
                 ts = ensure_aware(run.created_at)
                 if ts <= self.since:
@@ -183,54 +216,110 @@ class GitHubActions(CISystem):
                     log.info("Found run %s", run.run_number)
                     self.register_build(ts, True)
                     if run_event in event_types:
-                        yield GHABuildLog.from_workflow_run(self.dl_session, wf, run)
+                        event_id = self.get_event_id(run, run_event)
+                        yield GHABuildLog.from_workflow_run(
+                            self.dl_session, wf, run, run_event, event_id
+                        )
+                        if artifacts:
+                            for name, download_url in self.get_artifacts(run):
+                                yield GHAArtifact.from_workflow_run(
+                                    self.dl_session,
+                                    wf,
+                                    run,
+                                    run_event,
+                                    event_id,
+                                    name,
+                                    download_url,
+                                )
                     else:
                         log.info("Event type is %r; skipping", run.event)
 
+    def get_event_id(self, run: WorkflowRun, event_type: EventType) -> str:
+        if event_type is EventType.CRON:
+            return ensure_aware(run.created_at).strftime("%Y%m%dT%H%M%S")
+        elif event_type is EventType.PUSH:
+            return run.head_branch
+        elif event_type is EventType.PULL_REQUEST:
+            if run.pull_requests:
+                return str(run.pull_requests[0].number)
+            else:
+                # The experimental "List pull requests associated with a
+                # commit" endpoint does not return data reliably enough to be
+                # worth using, so we have to do an issue search for the
+                # matching PR instead.
+                try:
+                    return str(
+                        self.client.search_issues(
+                            f"repo:{run.repository.full_name} is:pr {run.head_sha}",
+                            sort="created",
+                            order="asc",
+                        )[0].number
+                    )
+                except IndexError:
+                    return "UNK"
+        else:
+            raise AssertionError(f"Unhandled EventType: {event_type!r}")
 
-@dataclass
-class GHABuildLog(BuildLog):
+    def get_artifacts(self, run: WorkflowRun) -> Iterator[Tuple[str, str]]:
+        """ Yields each artifact as a (name, download_url) pair """
+        url = run.artifacts_url
+        while url is not None:
+            r = self.dl_session.get(url)
+            r.raise_for_status()
+            for artifact in r.json()["artifacts"]:
+                if not artifact["expired"]:
+                    yield (artifact["name"], artifact["archive_download_url"])
+            url = r.links.get("next", {}).get("url")
+
+    def get_release_assets(self) -> Iterator["GHReleaseAsset"]:
+        log.info("Fetching releases newer than %s", self.since)
+        for rel in self.ghrepo.get_releases():
+            if rel.draft:
+                log.info("Release %s is draft; skipping", rel.tag_name)
+                continue
+            if rel.prerelease:
+                log.info("Release %s is prerelease; skipping", rel.tag_name)
+                continue
+            ts = ensure_aware(rel.published_at)
+            if ts <= self.since:
+                continue
+            self.register_build(ts, True)  # TODO: Set to False for drafts?
+            log.info("Found release %s", rel.tag_name)
+            r = self.dl_session.get(
+                f"https://api.github.com/repos/{self.repo}/git/refs/tags/{rel.tag_name}"
+            )
+            r.raise_for_status()
+            tagobj = r.json()["object"]
+            if tagobj["type"] == "commit":
+                commit = tagobj["sha"]
+            elif tagobj["type"] == "tag":
+                r = self.dl_session.get(tagobj["url"])
+                r.raise_for_status()
+                commit = r.json()["object"]["sha"]
+            else:
+                raise RuntimeError(
+                    f"Unexpected type for tag {rel.tag_name}: {tagobj['type']!r}"
+                )
+            for asset in rel.get_assets():
+                yield GHReleaseAsset(
+                    session=self.dl_session,
+                    published_at=ts,
+                    tag_name=rel.tag_name,
+                    commit=commit,
+                    name=asset.name,
+                    download_url=asset.browser_download_url,
+                )
+
+
+class GHAAsset(Asset):
     session: requests.Session
-    logs_url: str
     workflow_name: str
     workflow_file: str
     run_id: int
 
-    @classmethod
-    def from_workflow_run(
-        cls,
-        session: requests.Session,
-        workflow: Workflow,
-        run: WorkflowRun,
-    ) -> "GHABuildLog":
-        run_event = EventType.from_gh_event(run.event)
-        if run_event is None:
-            raise ValueError(f"Run has unknown event type {run.event!r}")
-        event_id: str
-        if run_event is EventType.CRON:
-            event_id = ensure_aware(run.created_at).strftime("%Y%m%dT%H%M%S")
-        elif run_event is EventType.PUSH:
-            event_id = run.head_branch
-        elif run_event is EventType.PULL_REQUEST:
-            if run.pull_requests:
-                event_id = str(run.pull_requests[0].number)
-            else:
-                event_id = "UNK"
-        else:
-            raise AssertionError(f"Unhandled EventType: {run_event!r}")
-        return cls(
-            session=session,
-            logs_url=run.logs_url,
-            created_at=ensure_aware(run.created_at),
-            event_type=run_event,
-            event_id=event_id,
-            commit=run.head_sha,
-            workflow_name=workflow.name,
-            workflow_file=workflow.path.split("/")[-1],
-            number=run.run_number,
-            run_id=run.id,
-            status=run.conclusion,
-        )
+    class Config:
+        # To allow requests.Session:
+        arbitrary_types_allowed = True
 
     def path_fields(self) -> Dict[str, str]:
         fields = super().path_fields()
@@ -244,6 +333,33 @@ class GHABuildLog(BuildLog):
         )
         return fields
 
+
+class GHABuildLog(GHAAsset, BuildLog):
+    logs_url: str
+
+    @classmethod
+    def from_workflow_run(
+        cls,
+        session: requests.Session,
+        workflow: Workflow,
+        run: WorkflowRun,
+        event_type: EventType,
+        event_id: str,
+    ) -> "GHABuildLog":
+        return cls(
+            session=session,
+            logs_url=run.logs_url,
+            created_at=ensure_aware(run.created_at),
+            event_type=event_type,
+            event_id=event_id,
+            commit=run.head_sha,
+            workflow_name=workflow.name,
+            workflow_file=workflow.path.split("/")[-1],
+            number=run.run_number,
+            run_id=run.id,
+            status=run.conclusion,
+        )
+
     def download(self, path: Path) -> List[Path]:
         path.mkdir(parents=True, exist_ok=True)
         if any(path.iterdir()):
@@ -263,6 +379,11 @@ class GHABuildLog(BuildLog):
             path,
         )
         r = self.session.get(self.logs_url)
+        if r.status_code == 404:
+            # This can happen when a workflow failed to run due to, say, a
+            # syntax error.
+            log.error("Request for logs returned 404; skipping")
+            return []
         r.raise_for_status()
         try:
             with BytesIO(r.content) as blob, ZipFile(blob) as zf:
@@ -273,6 +394,128 @@ class GHABuildLog(BuildLog):
         return list(path.rglob("*.txt"))
 
 
+class GHAArtifact(GHAAsset, Artifact):
+    name: str
+    download_url: str
+
+    @classmethod
+    def from_workflow_run(
+        cls,
+        session: requests.Session,
+        workflow: Workflow,
+        run: WorkflowRun,
+        event_type: EventType,
+        event_id: str,
+        name: str,
+        download_url: str,
+    ) -> "GHAArtifact":
+        return cls(
+            session=session,
+            created_at=ensure_aware(run.created_at),
+            event_type=event_type,
+            event_id=event_id,
+            commit=run.head_sha,
+            workflow_name=workflow.name,
+            workflow_file=workflow.path.split("/")[-1],
+            number=run.run_number,
+            run_id=run.id,
+            status=run.conclusion,
+            name=name,
+            download_url=download_url,
+        )
+
+    def download(self, path: Path) -> List[Path]:
+        target_dir = path / self.name
+        target_dir.mkdir(parents=True, exist_ok=True)
+        if any(target_dir.iterdir()):
+            log.info(
+                "Asset %s from %s (%s) #%s already downloaded to %s; skipping",
+                self.name,
+                self.workflow_file,
+                self.workflow_name,
+                self.number,
+                path,
+            )
+            return []
+        log.info(
+            "Downloading asset %s for %s (%s) #%s to %s",
+            self.name,
+            self.workflow_file,
+            self.workflow_name,
+            self.number,
+            path,
+        )
+        r = self.session.get(self.download_url, stream=True)
+        r.raise_for_status()
+        fd, fpath = tempfile.mkstemp()
+        os.close(fd)
+        zippath = Path(fpath)
+        try:
+            stream_to_file(r, zippath)
+            try:
+                with ZipFile(zippath) as zf:
+                    zf.extractall(target_dir)
+            except BaseException:
+                rmtree(target_dir)
+                raise
+        finally:
+            zippath.unlink(missing_ok=True)
+        return list(iterfiles(target_dir))
+
+
+class GHReleaseAsset(BaseModel):
+    session: requests.Session
+    published_at: datetime
+    tag_name: str
+    commit: str
+    name: str
+    download_url: str
+
+    class Config:
+        # To allow requests.Session:
+        arbitrary_types_allowed = True
+
+    def path_fields(self) -> Dict[str, str]:
+        utc_date = self.published_at.astimezone(timezone.utc)
+        return {
+            "year": utc_date.strftime("%Y"),
+            "month": utc_date.strftime("%m"),
+            "day": utc_date.strftime("%d"),
+            "hour": utc_date.strftime("%H"),
+            "minute": utc_date.strftime("%M"),
+            "second": utc_date.strftime("%S"),
+            "ci": "github",
+            "type": "release",
+            "type_id": self.tag_name,
+            "commit": self.commit,
+            "abbrev_commit": self.commit[:7],
+        }
+
+    def expand_path(self, path_template: str, vars: Dict[str, str]) -> str:
+        return expand_template(path_template, self.path_fields(), vars)
+
+    def download(self, path: Path) -> List[Path]:
+        target = path / self.name
+        if target.exists():
+            log.info(
+                "Asset %s for release %s already downloaded to %s; skipping",
+                self.name,
+                self.tag_name,
+                target,
+            )
+            return []
+        path.mkdir(parents=True, exist_ok=True)
+        log.info(
+            "Downloading asset %s for release %s to %s",
+            self.name,
+            self.tag_name,
+            target,
+        )
+        r = self.session.get(self.download_url, stream=True)
+        stream_to_file(r, target)
+        return [target]
+
+
 class Travis(CISystem):
     @staticmethod
     def get_auth_token() -> str:
@@ -327,7 +570,9 @@ class Travis(CISystem):
                 break
             params = None
 
-    def get_build_logs(self, event_types: List[EventType]) -> Iterator["TravisJobLog"]:
+    def get_assets(
+        self, event_types: List[EventType], artifacts: bool = False  # noqa: U100
+    ) -> Iterator["Asset"]:
         log.info("Fetching builds newer than %s", self.since)
         for build in self.paginate(
             f"/repo/{quote(self.repo, safe='')}/builds",
@@ -357,7 +602,6 @@ class Travis(CISystem):
                     log.info("Event type is %r; skipping", build["event_type"])
 
 
-@dataclass
 class TravisJobLog(BuildLog):
     client: Travis
     job: str
@@ -426,7 +670,6 @@ class TravisJobLog(BuildLog):
         return [path]
 
 
-@dataclass
 class Appveyor(CISystem):
     accountName: str
     projectSlug: Optional[str]
@@ -473,9 +716,9 @@ class Appveyor(CISystem):
             else:
                 break
 
-    def get_build_logs(
-        self, event_types: List[EventType]
-    ) -> Iterator["AppveyorJobLog"]:
+    def get_assets(
+        self, event_types: List[EventType], artifacts: bool = False  # noqa: U100
+    ) -> Iterator["Asset"]:
         log.info("Fetching runs newer than %s", self.since)
         for build in self.get_builds():
             if build.get("pullRequestId"):
@@ -501,7 +744,6 @@ class Appveyor(CISystem):
                     log.info("Event type is %r; skipping", run_event.value)
 
 
-@dataclass
 class AppveyorJobLog(BuildLog):
     client: Appveyor
     job: str
@@ -582,7 +824,9 @@ class CIConfig(NoExtraModel, ABC):
 
 
 class GitHubConfig(CIConfig):
-    workflows: List[str]
+    artifacts_path: Optional[str] = None
+    releases_path: Optional[str] = None
+    workflows: Optional[List[str]] = None
 
     @staticmethod
     def get_auth_token() -> str:
@@ -649,6 +893,7 @@ class DataladConfig(NoExtraModel):
 
 class Config(NoExtraModel):
     repo: str
+    vars: Dict[str, str] = Field(default_factory=dict)
     ci: CIConfigDict
     since: datetime
     types: List[EventType]
@@ -657,13 +902,13 @@ class Config(NoExtraModel):
     datalad: DataladConfig = Field(default_factory=DataladConfig)
 
     @validator("repo")
-    def _validate_repo(cls, v: str) -> str:  # noqa: B902
+    def _validate_repo(cls, v: str) -> str:  # noqa: B902, U100
         if not re.fullmatch(r"[^/]+/[^/]+", v):
             raise ValueError("Repo must be in the form 'OWNER/NAME'")
         return v
 
     @validator("since")
-    def _validate_since(cls, v: datetime) -> datetime:  # noqa: B902
+    def _validate_since(cls, v: datetime) -> datetime:  # noqa: B902, U100
         if v.tzinfo is None:
             raise ValueError("'since' timestamp must include timezone offset")
         return v
@@ -730,36 +975,60 @@ def fetch(cfg: Config, state: str, sanitize_secrets: bool) -> None:
     ds = Dataset(os.curdir)
     if cfg.datalad.enabled and not ds.is_installed():
         ds.create(force=True, cfg_proc=cfg.datalad.cfg_proc)
-    added = 0
+    logs_added = 0
+    artifacts_added = 0
+    relassets_added = 0
     for name, cicfg in cfg.ci.items():
-        log.info("Fetching logs from %s", name)
+        get_artifacts = getattr(cicfg, "artifacts_path", None) is not None
+        log.info("Fetching resources from %s", name)
         try:
             since = datetime.fromisoformat(since_stamps[name])
         except KeyError:
             since = cfg.since
         ci = cicfg.get_system(repo=cfg.repo, since=since, token=tokens[name])
-        for bl in ci.get_build_logs(cfg.types):
-            path = bl.expand_path(cicfg.path)
+        for obj in ci.get_assets(cfg.types, artifacts=get_artifacts):
+            if isinstance(obj, BuildLog):
+                path = obj.expand_path(cicfg.path, cfg.vars)
+            elif isinstance(obj, Artifact):
+                assert get_artifacts
+                path = obj.expand_path(
+                    cicfg.artifacts_path,  # type: ignore[attr-defined]
+                    cfg.vars,
+                )
+            else:
+                raise AssertionError(f"Unexpected asset type {type(obj).__name__}")
             if cfg.datalad.enabled:
-                dspaths = path.split("//")
-                if "" in dspaths:
-                    raise click.UsageError("Path contains empty '//'-delimited segment")
-                for i in range(1, len(dspaths)):
-                    dsp = "/".join(dspaths[:i])
-                    if not Path(dsp).exists():
-                        ds.create(dsp, cfg_proc=cfg.datalad.cfg_proc)
-            logs = bl.download(Path(path))
-            added += len(logs)
-            if sanitize_secrets and cfg.secrets:
-                for p in logs:
-                    sanitize(p, cfg.secrets, cfg.allow_secrets_regex)
+                ensure_datalad(ds, path, cfg.datalad.cfg_proc)
+            paths = obj.download(Path(path))
+            if isinstance(obj, BuildLog):
+                logs_added += len(paths)
+                if sanitize_secrets and cfg.secrets:
+                    for p in paths:
+                        sanitize(p, cfg.secrets, cfg.allow_secrets_regex)
+            elif isinstance(obj, Artifact):
+                artifacts_added += len(paths)
+        if isinstance(cicfg, GitHubConfig) and cicfg.releases_path is not None:
+            assert isinstance(ci, GitHubActions)
+            for asset in ci.get_release_assets():
+                path = asset.expand_path(cicfg.releases_path, cfg.vars)
+                if cfg.datalad.enabled:
+                    ensure_datalad(ds, path, cfg.datalad.cfg_proc)
+                paths = asset.download(Path(path))
+                relassets_added += len(paths)
         since_stamps[name] = ci.new_since().isoformat()
         log.debug("%s timestamp floor updated to %s", name, since_stamps[name])
     with open(state, "w") as fp:
         json.dump(since_stamps, fp)
-    log.info("%d logs downloaded", added)
-    if cfg.datalad.enabled and added:
-        ds.save(recursive=True, message=f"[tinuous] {added} logs added")
+    log.info("%d logs downloaded", logs_added)
+    log.info("%d artifacts downloaded", artifacts_added)
+    log.info("%d release assets downloaded", relassets_added)
+    if cfg.datalad.enabled and (logs_added or artifacts_added or relassets_added):
+        msg = f"[tinuous] {logs_added} logs added"
+        if artifacts_added:
+            msg += f", {artifacts_added} artifacts added"
+        if relassets_added:
+            msg += f", {relassets_added} release assets added"
+        ds.save(recursive=True, message=msg)
 
 
 @main.command("sanitize")
@@ -817,5 +1086,26 @@ def sanitize(
             fp.write(line)
 
 
+def iterfiles(dirpath: Path) -> Iterator[Path]:
+    dirs = deque([dirpath])
+    while dirs:
+        d = dirs.popleft()
+        for p in d.iterdir():
+            if p.is_dir():
+                dirs.append(p)
+            else:
+                yield p
+
+
+def ensure_datalad(ds: Dataset, path: str, cfg_proc: Optional[str]) -> None:
+    dspaths = path.split("//")
+    if "" in dspaths:
+        raise click.UsageError("Path contains empty '//'-delimited segment")
+    for i in range(1, len(dspaths)):
+        dsp = "/".join(dspaths[:i])
+        if not Path(dsp).exists():
+            ds.create(dsp, cfg_proc=cfg_proc)
+
+
 if __name__ == "__main__":
     main()
